﻿function Decoder(bytes, port) {
    // Beim Testen bytes ausgeben
    console.log(bytes);

    //https://stackoverflow.com/questions/37471158/converting-ieee-754-from-bit-stream-into-float-in-javascript/37471538#37471538 (Bursov, May 26, 2016)
    // we can't use DataView etc. here because ChirpStack uses https://github.com/robertkrimen/otto which targets ES5
    function bytesToFloat(bytesToDecode) {

        // Assume LSB (least significant byte first).
        var bits = bytesToDecode[0] << 24 | bytesToDecode[1] << 16 | bytesToDecode[2] << 8 | bytesToDecode[3];

        var sign = (bits >>> 31 === 0) ? 1.0 : -1.0;
        var e = bits >>> 23 & 0xff;
        var m = (e === 0) ? (bits & 0x7fffff) << 1 : (bits & 0x7fffff) | 0x800000;
        if (e == 0xff) {
            if ((bits & 0x7fffff) == 0) {
                return sign / 0;
            } else {
                return NaN;
            }
        }
        var f = sign * m * Math.pow(2, e - 150);
        return f;
    }

    function bytesToInt16(bytesToDecode) {
        var tmpInt16 = (bytesToDecode[0] << 24 >> 16) | bytesToDecode[1];
        return tmpInt16;
    }

    function bytesToUInt16(bytesToDecode) {
        var tmpInt16 = bytesToDecode[0] * 256 + bytesToDecode[1];
        return tmpInt16;
    }

    function bytesToUInt32(bytesToDecode) {
        var tmpUInt32 = bytesToDecode[0] * 256 * 256 * 256 + bytesToDecode[1] * 256 * 256 + bytesToDecode[2] * 256 + bytesToDecode[3];
        return tmpUInt32;
    }

    function bytesToString(bytesToDecode) {
        var str = '';
        for (var i = 0; i < bytesToDecode.length; i++) {
            var chr = bytesToDecode[i];
            if (chr == 0)
                break;
            str += String.fromCharCode(chr);
        }
        return str;
    }

    var decoded = {};
    decoded.PacketIdentifier = bytes[0];
    if (decoded.PacketIdentifier == 2 || decoded.PacketIdentifier == 3 || decoded.PacketIdentifier == 4 || decoded.PacketIdentifier == 5) {
        decoded.NamurState = bytes[1];
        decoded.Distance = bytesToFloat(bytes.slice(2, 6));
        decoded.Unit = bytes[6];
        decoded.RemainingPower = bytes[7];
        decoded.Temperature = bytesToInt16(bytes.slice(8, 10)) / 10;
        decoded.TemperatureUnit = 32; // HartUnitCode DegreeC

        if (decoded.PacketIdentifier == 2) {
            decoded.Inclination_degree = bytes[10];
        }
        else if (decoded.PacketIdentifier == 3) {
            decoded.GNSSLatitude = bytesToFloat(bytes.slice(10, 14));
            decoded.GNSSLongitude = bytesToFloat(bytes.slice(14, 18));
            decoded.Inclination_degree = bytes[18];
        }
        else if (decoded.PacketIdentifier == 4) {
            decoded.DetailState = bytesToUInt32(bytes.slice(10, 14));
            decoded.Inclination_degree = bytes[14];
        }
        else if (decoded.PacketIdentifier == 5) {
            decoded.GNSSLatitude = bytesToFloat(bytes.slice(10, 14));
            decoded.GNSSLongitude = bytesToFloat(bytes.slice(14, 18));
            decoded.DetailState = bytesToUInt32(bytes.slice(18, 22));
            decoded.Inclination_degree = bytes[22];
        }
    }
    else if (decoded.PacketIdentifier == 6) {
        decoded.NamurState = bytes[1];
        decoded.GNSSLatitude = bytesToFloat(bytes.slice(2, 6));
        decoded.GNSSLongitude = bytesToFloat(bytes.slice(6, 10));
    }
    else if (decoded.PacketIdentifier == 7) {
        decoded.NamurState = bytes[1];
        decoded.DetailState = bytesToUInt32(bytes.slice(2, 6));
    }
    else if ((decoded.PacketIdentifier >= 8 && decoded.PacketIdentifier <= 15) || (decoded.PacketIdentifier >= 18 && decoded.PacketIdentifier <= 22)) {
        var position = 1; // skip identifier
        if ([10, 11, 14, 15, 19, 20, 21, 22].indexOf(decoded.PacketIdentifier) != -1) {
            decoded.NamurState = bytes[position];
            position++;
        } else {
            decoded.NamurState = 0;
        }

        if ([8, 9, 10, 11, 12, 13, 14, 15, 18, 19].indexOf(decoded.PacketIdentifier) != -1) {
            decoded.Distance = bytesToFloat(bytes.slice(position, position + 4));
            position += 4;
            decoded.Unit = bytes[position];
            position++;
        }

        if ([12, 13, 14, 15, 22].indexOf(decoded.PacketIdentifier) != -1) {
            if (bytes[position] == 0x80 && bytes[position + 1] == 0x00) {
                decoded.PercentValue = NaN;
            }
            else {
                decoded.PercentValue = bytesToInt16(bytes.slice(position, position + 2)) / 100;
            }
            position += 2;

            if (bytes[position] == 0x80 && bytes[position + 1] == 0x00) {
                decoded.LinPercentValue = NaN;
            }
            else {
                decoded.LinPercentValue = bytesToInt16(bytes.slice(position, position + 2)) / 100;
            }
            position += 2;

            decoded.ScaledValue = bytesToFloat(bytes.slice(position, position + 4));
            position += 4;
            decoded.ScaledValueUnit = bytes[position];
            position++;
        }

        if ([8, 9, 10, 11, 12, 13, 14, 15, 18, 19].indexOf(decoded.PacketIdentifier) != -1) {
            decoded.RemainingPower = bytes[position];
            position++;
        }

        if ([9, 11, 13, 15, 21].indexOf(decoded.PacketIdentifier) != -1) {
            decoded.GNSSLatitude = bytesToFloat(bytes.slice(position, position + 4));
            if (isNaN(decoded.GNSSLatitude) && (bytes[position] & 0x80) > 0) { // NaN with sign bit set = GPS not attempted, otherwise GPS failed
                delete decoded.GNSSLatitude;
            }
            position += 4;

            decoded.GNSSLongitude = bytesToFloat(bytes.slice(position, position + 4));
            if (isNaN(decoded.GNSSLongitude) && (bytes[position] & 0x80) > 0) { // NaN with sign bit set = GPS not attempted, otherwise GPS failed
                delete decoded.GNSSLongitude;
            }
            position += 4;
        }

        if ([10, 11, 14, 15, 20].indexOf(decoded.PacketIdentifier) != -1) {
            decoded.DetailState = bytesToUInt32(bytes.slice(position, position + 4));
            position += 4;
        }

        if ([8, 9, 10, 11, 12, 13, 14, 15, 18, 20].indexOf(decoded.PacketIdentifier) != -1) {
            if (bytes[position] == 0x80 && bytes[position + 1] == 0x00) {
                decoded.Temperature = NaN;
            }
            else {
                decoded.Temperature = bytesToInt16(bytes.slice(position, position + 2)) / 10;
            }
            position += 2;

            decoded.TemperatureUnit = bytes[position];
            position++;
        }

        if ([8, 9, 10, 11, 12, 13, 14, 15, 18, 19].indexOf(decoded.PacketIdentifier) != -1) {
            decoded.Inclination_degree = bytes[position];
            position++;
        }
    }
    else if ([16, 17, 23, 24, 25, 26].indexOf(decoded.PacketIdentifier) != -1) {
        var position = 1; // skip identifier

        if ([23, 24, 25, 26].indexOf(decoded.PacketIdentifier) != -1) {
            decoded.NamurState = bytes[position];
            position++;
        }

        if ([16, 23].indexOf(decoded.PacketIdentifier) != -1) {
            decoded.Information = bytes[position];
            position++;
        }

        if ([16, 23].indexOf(decoded.PacketIdentifier) != -1) {
            decoded.DtmId = bytesToUInt32(bytes.slice(position, position + 4));
            position += 4;
            decoded.ManufacturerId = bytesToUInt32(bytes.slice(position, position + 4));
            position += 4;
        }

        if ([16, 24].indexOf(decoded.PacketIdentifier) != -1) {
            decoded.DeviceType = bytesToUInt32(bytes.slice(position, position + 4));
            position += 4;

            decoded.SoftwareVersionAscii = bytes[position] + '.' + bytes[position + 1] + '.' + bytes[position + 2] + '.' + bytes[position + 3];
            position += 4;
        }

        if ([16, 25].indexOf(decoded.PacketIdentifier) != -1) {
            var schedule = {};

            // Days (bits 0-7 (LSB = 0))
            var daysRaw = bytes[position] >>> 1 & 0x7f;
            var days = ['Su', 'Mo', 'Tu', 'We', 'Th', 'Fr', 'Sa'];
            schedule.Days = [];
            for (var i = 0; i < 7; i++) {
                if ((daysRaw & (1 << i)) > 0) {
                    schedule.Days.push(days[i]);
                }
            }

            //StartTimeMinutes (bits 8-18)
            schedule.StartTimeMinutes = ((bytes[position] & 0x01) << 10) | (bytes[position + 1] << 2) | ((bytes[position + 2] & 0xC0) >>> 6);

            //EndTimeMinutes (bits 19-29)
            schedule.EndTimeMinutes = ((bytes[position + 2] & 0x3F) << 5) | ((bytes[position + 3] & 0xF8) >>> 3);

            //MeasureIntervalMinutes (bits 30-40)
            schedule.MeasureIntervalMinutes = ((bytes[position + 3] & 0x07) << 8) | bytes[position + 4];

            //TransmissionIntervalMinutes (bits 46-56)
            schedule.TransmissionIntervalMinutes = ((bytes[position + 5] & 0x07) << 8) | bytes[position + 6];

            decoded.Schedule = schedule;
            position += 7;

            decoded.ChangeCounter = bytesToUInt16(bytes.slice(position, position + 2));
            position += 2;
        }

        if ([16, 26].indexOf(decoded.PacketIdentifier) != -1) {
            decoded.ScaledMin = bytesToFloat(bytes.slice(position, position + 4));
            position += 4;

            decoded.ScaledMax = bytesToFloat(bytes.slice(position, position + 4));
            position += 4;
        }

        if (decoded.PacketIdentifier == 17) {
            decoded.DeviceName = bytesToString(bytes.slice(position, position + 19));
            position += 19;

            decoded.DeviceTag = bytesToString(bytes.slice(position, position + 19));
            position += 19;
        }
    }

    return decoded;
}